package background;

import java.util.ArrayList;

import background.commands.Define;
import background.expressions.MathTree;
import background.systemVariables.ArrayVariable;
import background.systemVariables.BooleanVariable;
import background.systemVariables.CharacterVariable;
import background.systemVariables.IntegerVariable;
import background.systemVariables.PointerVariable;
import background.systemVariables.SimpleVariable;
import background.systemVariables.StringVariable;
import background.systemVariables.TableVariable;

public class VariableManager {

	public ArrayList<SimpleVariable> variables;
	public ArrayList<ArrayVariable> arrays;
	public ArrayList<TableVariable> tables;
	public ArrayList<SimpleVariable> actives;	
	public ArrayList<PointerVariable> pointers;
	
	public ArrayList<String> usedNames;
	public AddressSpace space;
	
	public ErrorHandler errorHandler;
	
	public VariableManager(ErrorHandler eh) {
		errorHandler = eh;
		reset();
	}
	
	public void reset() {
		variables = new ArrayList<SimpleVariable>();
		arrays = new ArrayList<ArrayVariable>();
		tables = new ArrayList<TableVariable>();
		actives = new ArrayList<SimpleVariable>();
		usedNames = new ArrayList<String>();		
		pointers = new ArrayList<PointerVariable>();
		space = new AddressSpace(this);
	}
	
	public void resetActives() {
		for (SimpleVariable var: actives) {
			var.setInactive();			
		}
		actives = new ArrayList<SimpleVariable>();
	}
	
	public void makeActive(SimpleVariable s) {
		s.setActive();
		actives.add(s);
	}

	
	/* ======================================================================================== */
	
	public void addVariable(String name, int value) {
		this.addVariable(name, new IntegerVariable(value));
	}
	
	public void addVariable(String name, String value) {
		this.addVariable(name, new StringVariable(value, this));
	}

	public void addVariable(String name, Boolean value) {
		this.addVariable(name, new BooleanVariable(value));
	}

	public void addVariable(String name, char value) {
		this.addVariable(name, new CharacterVariable(value));			
	}	
	
	public void addVariable(String name, String type, String value, Boolean vis) {	
		if (type.equals("integer")) type = "int";	
		if (value==null) { value = getDefaultValueForType(type); }
		SimpleVariable x = getNewVariableOfType(type, value); System.out.println("GET TYPE: " + type);
		x.setVizualize(vis);
		this.addVariable(name, x);		
	}
	

	public void addVariable(String name, String type, String value) {
		if (value==null) { value = getDefaultValueForType(type); }
		this.addVariable(name, getNewVariableOfType(type, value));		
	}
	
	public void addVariable(String name, SimpleVariable var) {
		if (usedNames.contains(name)) errorHandler.errorHappened("UsedVarName");
		else {
			var.name = name;
			variables.add(var);	
			usedNames.add(name);
			makeActive(var);
			space.add(name, 1, 0);
		}
	}
	

	public void addPointer(String name, String type, String value, Boolean vis) {
		if (value==null) { value = "0"; }
		PointerVariable x = new PointerVariable(name, type, value, this);
		x.setVizualize(vis);
		
		if (usedNames.contains(name)) errorHandler.errorHappened("UsedVarName");
		else {
			x.name = name;
			pointers.add(x);	
			usedNames.add(name);
			makeActive(x);
			space.add(name, 1, 0);
		}	
	}
	
	public void addArray(String name, int size, String type, Boolean pointer, Define d) {
		addArray(name, 0, size, type, false, pointer, d);
	} 
	
	public void addArray(String name, int size, String type, Boolean vis, Boolean pointer, Define d) {
		addArray(name, 0, size, type, vis, pointer, d);
	} 
	
	public void addArray(String name, int posun, int size, String type, Boolean vis, Boolean pointer, Define define) {
		if (usedNames.contains(name)) errorHandler.errorHappened("UsedVarName"); 
		else {
			ArrayVariable a = new ArrayVariable(name, posun, size, type, pointer, define, this);
			a.name = name;
			a.vizualize = vis;
			a.pointer = pointer;
			arrays.add(a);
			usedNames.add(name);
			space.add(name, size, posun);
		}
	} 

	
	public void addTable(String name, int size1, int size2, String type, Boolean pointer, Define define) {
		addTable(name, 0, size1, 0, size2, type, false, pointer, define);
	} 
	
	public void addTable(String name, int size1, int size2, String type, Boolean vis, Boolean pointer, Define define) {		
		addTable(name, 0, size1, 0, size2, type, vis, pointer, define);
	} 

	public void addTable(String name, int posun1, int size1, int posun2, int size2, String type, Boolean vis, Boolean pointer, Define define) {
		if (usedNames.contains(name)) errorHandler.errorHappened("UsedVarName");
		else {
			TableVariable a = new TableVariable(name, posun1, size1, posun2, size2, type, pointer, define, this);
			a.name = name;
			a.vizualize = vis;
			a.pointer = pointer;
			tables.add(a);
			usedNames.add(name);
			space.add(name, size1, size2, posun1, posun2);
		}
	} 

	
	/* ======================================================================================== */
	
	public void updateVariable(String name, String newValue) {		
		SimpleVariable a = getVariableByName(name);
		
		if (name.contains(".s") && !(a instanceof PointerVariable)) {
			String parts[] = name.split(".s");
			if (parts.length > 0) {
				SimpleVariable b = getVariableByName(parts[0]);
				b.initialized = true;
				System.out.println("MAKE ACTIVE " + parts[0]);
				makeActive(b);
				a.initialized = true;
				a.setValue(newValue);
				makeActive(a);
			}
		}
		else if (a instanceof PointerVariable) updatePointer(name, newValue);
		else {
			a.setValue(newValue);
			a.initialized = true;
			makeActive(a);
		}
	}
	

	public void updatePointer(String name, String value) {
		PointerVariable a = (PointerVariable) getVariableByName(name);
		if (a.pointer != null) {
			a.setValue(value);
			a.initialized = true;
			makeActive(a);
			if (a.pointer != null) makeActive(a.pointer);
		} else {
			errorHandler.errorHappened("Initialized");
		}
	}

	public void updateArray(String name, ArrayVariable a) {
		arrays.remove(getArrayVariableByName(name));
		arrays.add(a);		
	}
	
	public void updateTable(String name, TableVariable a) {
		tables.remove(getTableVariableByName(name));
		tables.add(a);		
	}
	
	public void updateArrayVariable(String name, int index, String newValue) {
		ArrayVariable a = getArrayVariableByName(name);
		
		if ((index>=a.size+a.posun) || (index<a.posun)) { errorHandler.errorHappened("Index");}
		else {
			a.updateNth(index, this.getNewVariableOfType(a.getType(), newValue));
			makeActive(a.getNth(index));
		}	
	}
	
	public void updateTableVariable(String name, int index, int index2, String newValue) {
		TableVariable a = getTableVariableByName(name);
				
		if ((index2>=a.sizeX+a.posunX) || (index2<a.posunX) || (index>=a.sizeY+a.posunY) || (index<a.posunY)) { errorHandler.errorHappened("Index");}
		else { a.updateNth(index, index2, this.getNewVariableOfType(a.getType(), newValue));
			makeActive(a.getNth(index, index2));
		}
	}

	
	/* ======================================================================================== */

	
	public ArrayList<SimpleVariable> getAllVariables() {
		return variables;
	}
	
	public ArrayList<ArrayVariable> getAllArrays() {
		return arrays;
	}	
	
	public ArrayList<TableVariable> getAllTables() {
		return tables;
	}
	
	public ArrayList<PointerVariable> getAllPointers() {
		return pointers;
	}
	
	public ArrayList<SimpleVariable> getVizualizedVariables() {
		ArrayList<SimpleVariable> vars = new ArrayList<SimpleVariable>();
		for (SimpleVariable var: variables) {
			if (var.vizualize) {vars.add(var);}
		}		
		return vars;
	}

	public ArrayList<SimpleVariable> getActiveVisibleVariables() {
		ArrayList<SimpleVariable> vars = new ArrayList<SimpleVariable>();
		for (SimpleVariable var: variables) {
			if (var.vizualize && var.isActive()) {vars.add(var);}
		}		
		return vars;
	}

	
	public ArrayList<ArrayVariable> getVizualizedArrays() {
		ArrayList<ArrayVariable> vars = new ArrayList<ArrayVariable>();
		for (ArrayVariable var: arrays) {
			if (var.vizualize) { vars.add(var); }
		}		
		return vars;
	}
	
	public ArrayList<TableVariable> getVizualizedTables() {
		ArrayList<TableVariable> vars = new ArrayList<TableVariable>();
		for (TableVariable var: tables) {
			if (var.vizualize) { vars.add(var); }
		}		
		return vars;
	}
	
	
	/* ======================================================================================== */
	
	public SimpleVariable getNewVariableOfType(String type, String value) {
		switch (type.toLowerCase()) {
			case "Integer":
			case "int": return new IntegerVariable(value);
			case "string": return new StringVariable(value, this);
			case "char": return new CharacterVariable(value);
			case "boolean": return new BooleanVariable(value);
			default: return null;
		}
	}
	
	public SimpleVariable getVariableByName(String name) {	
		
		// musime prezriet adresny priestor
		if (name.contains("@")) {
			return space.getPosition(name);
		}
		
		// ak chceme adresu kam ukazuje pointer
		if (name.contains("$")) {
			SimpleVariable pointer = getVariableByName(name.substring(1));
			if (pointer != null) return space.getPosition((((PointerVariable) pointer).expression));
			else return null;
		}
		
		if (name.contains("[") && name.contains("]")) {
			String arrayV[] = name.split("[\\[\\]]");
			if (arrayV.length>2) {
				for (TableVariable variable: this.tables) {
					if (variable.name.equals(arrayV[0])) {											// najdem spravne pole						
						MathTree mt = new MathTree(arrayV[1], this);
						int index = mt.getTreeValue().intValue();
						MathTree mt2 = new MathTree(arrayV[3], this);
						int index2 = mt2.getTreeValue().intValue();						
						if ((index2>=variable.sizeY+variable.posunY) || (index2<variable.posunY) || (index>=variable.sizeX+variable.posunX) || (index<variable.posunX)) { errorHandler.errorHappened("Index"); return null;}
						else return variable.getNth(index2, index);					// vyberiem z neho tu premennu ktora je na mieste urcenom premennou v zatvorkach
					}
				}
			}	
			
			
			for (ArrayVariable variable: this.arrays)
				if (variable.name.equals(arrayV[0])) {											// najdem spravne pole
			
					MathTree mt = new MathTree(arrayV[1], this);
					int index = mt.getTreeValue().intValue();
					
					if ((index>=variable.size+variable.posun) || (index<variable.posun)) { errorHandler.errorHappened("Index"); return null;}
					else {
						if ((arrayV.length>2) && variable.getType().equals("string")) {
							MathTree mt2 = new MathTree(arrayV[3], this);
							int index2 = mt2.getTreeValue().intValue();
							StringVariable str = (StringVariable) variable.getNth(index);
							if ((index2>str.getValue("string").length()) || (index2<0)) { errorHandler.errorHappened("Index"); return null;}
							else return str.getCharAt(index2);					// vyberiem z neho tu premennu ktora je na mieste urcenom premennou v zatvorkach
						} else {																				
							return variable.getNth(index);					// vyberiem z neho tu premennu ktora je na mieste urcenom premennou v zatvorkach
						}
					}
				}					
			
			
			for (SimpleVariable variable: this.variables) {
				if (variable.name.equals(arrayV[0]) && variable.getType().equals("string")) {											// najdem spravne pole
					MathTree mt = new MathTree(arrayV[1], this);
					StringVariable x = (StringVariable) variable;
					int index = mt.getTreeValue().intValue(); 
					if ((index>x.getValue("string").length()) || (index<0)) { errorHandler.errorHappened("Index"); return null;}
					else return x.getCharAt(index);					// vyberiem z neho tu premennu ktora je na mieste urcenom premennou v zatvorkach
				}
			}
		} else if (name.contains("*") || name.contains("^")) { 
			name = name.replace("*", "").replace("^", "");
			
			for (PointerVariable variable : this.pointers) {
				if (variable.name.equals(name)) return variable;
			}
		} else {
			for (SimpleVariable variable: this.variables) {
				if (variable.name.equals(name)) return variable;
			}			
			for (SimpleVariable variable: this.pointers) {
				if (variable.name.equals(name)) return variable;
			}			
		}		
		
		errorHandler.errorHappened("VarNotFound");
		return null;
	}
		
	public ArrayVariable getArrayVariableByName(String name) {
		for (ArrayVariable variable: this.arrays) 
			if (variable.name.equals(name)) return variable;						
		return null;
	}
	
	public TableVariable getTableVariableByName(String name) {
		for (TableVariable variable: this.tables) 
			if (variable.name.equals(name)) return variable;						
		return null;
	}

	public PointerVariable getPointerVariableByName(String name) {
		for (PointerVariable variable: this.pointers) 
			if (variable.name.equals(name)) return variable;	
		return null;
	}

	public String getDefaultValueForType(String type) {
		switch (type.toLowerCase()) {
			case "int": return "0";
			case "char": return " ";
			case "string": return "";
			case "boolean": return "false";
		}
		return null;
	}
	
	public static ArrayList<String> getVariableTypes() {
		ArrayList<String> result = new ArrayList<String>();
		String[] types = new String[]{ "int","integer","char","string","boolean"};
		for (int i=0; i<types.length; i++) {
			result.add(types[i]);
		}
		return result;		
	}

	
	
	/* ===================================  DEBUG  ============================================= */
	
	public void writeAllVariables() {
		for (SimpleVariable var: variables) {
			writeVariable(var, true);
			System.out.print("   ");
		}
		System.out.println();
	}

	public void writeVisibleVariables() {
		for (SimpleVariable var: this.getVizualizedVariables()) {
			writeVariable(var, true);
			System.out.print("   ");
		}
		System.out.println();
	}

	
	public void writeAllArrays() {
		
		for (ArrayVariable a: arrays) {
			System.out.print(a.name + ": ");
			for (int i=a.posun; i<a.size+a.posun; i++) {
				System.out.print("[" + i + "] -> ");
				writeVariable(a.getNth(i), false);
				System.out.print("   ");
			}
			System.out.println();
		}	
	}

	public void writeAllTables() {
		
		for (TableVariable a: tables) {
			System.out.print(a.name + ": ");
			for (int i=a.posunX; i<a.sizeX+a.posunX; i++) {
				for (int j=a.posunY; j<a.sizeY+a.posunY; j++) {
					System.out.print("[" + i + "," +j+"] -> ");
					writeVariable(a.getNth(j, i), false);
					System.out.print("   ");
				}
				System.out.println();
			}
			System.out.println();
		}	
	}

	
	private void writeVariable(SimpleVariable v, Boolean showName) {
		if (showName && v.vizualize) System.out.print("*");
		if (!v.initialized) System.out.print("!");
		if (showName) System.out.print(v.name + " = ");
		switch (v.getType()) {
			case "int": System.out.print(v.getValue(1)); break;
			case "char": System.out.print("'"+v.getValue('c')+"'"); break;
			case "string": System.out.print("\""+v.getValue("string")+"\""); break;
			case "boolean": System.out.print(v.getValue(true)); break;
			case "": System.out.print(""); break;
		}
	}


	
	
}
